#include "Halide.h"
#include <cstdio>

using namespace Halide;

Target rvv_define_target(int vlen, int arch_bits) {
        Target target("host-no_asserts-no_runtime-no_bounds_query");
        target.os = Target::Linux;                // The operating system
        target.arch = Target::RISCV;                  // The CPU architecture
        target.bits = arch_bits;                           // The bit-width of the architecture
        target.vector_bits = vlen;                           
        std::vector<Target::Feature> rvv_features;  // A list of features to set
        rvv_features.push_back(Target::RVV);
        target.set_features(rvv_features);
        // printf(target)
    return target;
}

Target sve_define_target(int vlen, int arch_bits) {
    Target sve_target("host-no_asserts-no_runtime-no_bounds_query");
    sve_target.os = Target::Linux;                // The operating system
    sve_target.arch = Target::ARM;                  // The CPU architecture
    sve_target.bits = arch_bits;                           // The bit-width of the architecture
    sve_target.vector_bits = vlen;                           
    std::vector<Target::Feature> sve_features;  // A list of features to set
    sve_features.push_back(Target::SVE);
    sve_target.set_features(sve_features);
    return sve_target;
}

int main(int argc, char **argv) {


    printf("exceptions enabled: %d\n", exceptions_enabled());

    const int arch_bits = 64;
    const int vlen = 256;
    const int num_elems = vlen/16;
    const int matrix_size = num_elems*10;

    ImageParam A(type_of<uint8_t>(), 2);
    std::vector<Argument> args(1);
    args[0] = A;

    Var x("x"),  y("y");
    // Func blur_x("blur_x");
    Func out("out");
    Func rows{"rows"}, cols{"cols"}, bounded_input{"bounded_input"};

    //define blur function
    // printf("define blur \n");
    bounded_input(x, y) = BoundaryConditions::repeat_edge(A)(x, y);

    Func input_16("input_16");
    input_16(x, y) = cast<uint16_t>(bounded_input(x, y));

    // 5x5 Kernel
    // cols(x, y) = input_16(x - 2, y) + 4 * input_16(x - 1, y) + 6 * input_16(x, y) + 4 * input_16(x + 1, y) + input_16(x + 2, y);
    // rows(x, y) = cols(x, y - 2) + 4 * cols(x, y - 1) + 6 * cols(x, y) + 4 * cols(x, y + 1) + cols(x, y + 2);
    // out(x, y) = cast<uint8_t>(rows(x, y) >> 8);

    // 3x3 Kernel
    cols(x, y) =  input_16(x - 1, y) + 2 * input_16(x, y) + input_16(x + 1, y);
    rows(x, y) = cols(x, y - 1) + 2 * cols(x, y) + cols(x, y + 1);
    out(x, y) = cast<uint8_t>(rows(x, y) >> 4);

    //schedule
    Var xi("xi"), yi("yi");
    printf("define schedule \n");
    out//.split(y, y, yi, 8).parallel(y)
        .vectorize(x, num_elems);
        // .compute_at(out, x);

    printf("printing loop nest \n");
    out.print_loop_nest();
    
    // printf("defining SVE target \n");
    Target sve_target;
    sve_target = sve_define_target(vlen, arch_bits);  
    printf("compiling to SVE \n");
    out.compile_to_assembly("gaussian_blur.sve.S", args, "out", sve_target);
    printf("completed compile to SVE \n");


    // printf("defining RVV target \n");
    Target rvv_target;
    rvv_target = rvv_define_target(vlen, arch_bits);  
    printf("compiling to RVV \n");
    out.compile_to_assembly("gaussian_blur.rvv.S", args, "out", rvv_target);
    out.compile_to_static_library("gaussian_blur", args, "out", rvv_target);
    printf("completed compile to RVV \n");

    printf("Success!\n");
    return 0;
}
